"""
Parsers from JEOL.
"""

# Standard library modules.
import logging
import os
import pkgutil

# Third party modules.

# Local modules.
from pyxray.descriptor import Reference, Element, XrayTransition
from pyxray.property import XrayTransitionEnergy, XrayTransitionRelativeWeight
from pyxray.util import wavelength_to_energy_eV
import pyxray.parser.base as base

# Globals and constants variables.
logger = logging.getLogger(__name__)

JEOL = Reference("JEOL")

_TRANSITION_LOOKUP = {
    # Siegbahn
    "KA1": base.Ka1,
    "KA2": base.Ka2,
    "KB1": base.Kb1,
    "KB2": base.Kb2,
    "KB2_1": base.Kb2_1,
    "KB2+2": base.Kb2_2,
    "KB2_2": base.Kb2_2,
    "KB3": base.Kb3,
    "KB4": base.Kb4,
    "KB4_1": base.Kb4_1,
    "KB4_2": base.Kb4_2,
    "KB4x": base.Kb4_2,
    "KB5": base.Kb5,
    "KB5_1": base.Kb5_1,
    "KB5+2": base.Kb5_2,
    "KB5_2": base.Kb5_2,
    "LA1": base.La1,
    "LA2": base.La2,
    "LN": base.Ln,
    "LL": base.Ll,
    "LS": base.Ls,
    "LT": base.Lt,
    "LV": base.Lv,
    "LB1": base.Lb1,
    "LB2": base.Lb2,
    "LB3": base.Lb3,
    "LB4": base.Lb4,
    "LB6": base.Lb6,
    "LB7": base.Lb7,
    "LB9": base.Lb9,
    "LB10": base.Lb10,
    "LB15": base.Lb15,
    "LB17": base.Lb17,
    "LG1": base.Lg1,
    "LG2": base.Lg2,
    "LG3": base.Lg3,
    "LG4": base.Lg4,
    "LG4_p": base.Lg4p,
    "LG5": base.Lg5,
    "LG6": base.Lg6,
    "LG8": base.Lg8,
    "MA1": base.Ma1,
    "MA2": base.Ma2,
    "MB": base.Mb,
    "MG": base.Mg,
    "MZ1": base.Mz1,
    "MZ2": base.Mz2,
    # IUPAC
    "M1-N2": XrayTransition(base.N2, base.M1),
    "M1-N3": XrayTransition(base.N3, base.M1),
    "M2-M4": XrayTransition(base.M4, base.M2),
    "M2-N1": XrayTransition(base.N1, base.M2),
    "M2-N4": XrayTransition(base.N4, base.M2),
    "M2-O4": XrayTransition(base.O4, base.M2),
    "M3-M5": XrayTransition(base.M5, base.M3),
    "M3-N1": XrayTransition(base.N1, base.M3),
    "M3-N4": XrayTransition(base.N4, base.M3),
    "M4-O2": XrayTransition(base.O2, base.M4),
    "M4-O3": XrayTransition(base.O3, base.M4),
    "M5-N1": XrayTransition(base.N1, base.M5),
    "M5-O3": XrayTransition(base.O3, base.M5),
    "N6-O4": XrayTransition(base.O4, base.N6),
    "N7-O5": XrayTransition(base.O5, base.N7),
    # Set
    # left out transition sets: KBX, KB5+, L2,3-M
    "KA": base.Ka,
    "KA1,2": base.Ka,
    # 'KB': [(M3, K), (M2, K), (M5, K), (M4, K)], # FIXME: Not quite sure what to do with Kb
    "KB1,3": base.Kb1_3,
    "LA1,2": base.La,
    "LB2,15": base.Lb2_15,
    "LB5": base.Lb5,
    "LB3,4": base.Lb3_4,
    "LG2,3": base.Lg2_3,
    "MA": base.Ma,
    "MZ": base.Mz,
    "K-O2,3": base.KO2_3,
    "M1-N2,3": base.M1N2_3,
    "M2,3M4,5": base.M2_3M4_5,
    "M4,5O2,3": base.M4_5O2_3,
    "M3-O4,5": base.M3O4_5,
    "M4-O2,3": base.M4O2_3,
    "LL,N": base.Ll_n,
}


class JEOLTransitionParser(base._Parser):
    def __iter__(self):
        relpath = os.path.join("..", "data", "lambda.asc")
        content = pkgutil.get_data(__name__, relpath).decode("utf8")

        notread = set()
        transition_energy = []
        for line in content.splitlines():
            line = line.strip()
            if not line:
                continue

            z = int(line[0:2])

            siegbahn = line[10:18].strip()
            if siegbahn.startswith("A"):  # skip absorption edges
                continue
            if siegbahn.startswith("S"):  # skip satellite lines
                continue
            if siegbahn not in _TRANSITION_LOOKUP:  # check for equivalence
                notread.add(siegbahn)
                continue

            probability = line[20:23].strip()
            if not probability:  # skip transition with no probability
                continue
            probability = float(probability) / 100.0

            wavelength = float(line[26:35])
            energy = wavelength_to_energy_eV(wavelength * 1e-10)

            if siegbahn in _TRANSITION_LOOKUP:
                transition = _TRANSITION_LOOKUP[siegbahn]
                transition_energy.append((z, transition, probability, energy))
                continue

        length = len(transition_energy)
        for z, transition, probability, eV in transition_energy:
            if eV is None:
                continue
            element = Element(z)

            prop = XrayTransitionEnergy(JEOL, element, transition, eV)
            logger.debug("Parsed: {0}".format(prop))
            self.update(int((z - 1) / length * 100.0))
            yield prop

            prop = XrayTransitionRelativeWeight(JEOL, element, transition, probability)
            logger.debug("Parsed: {0}".format(prop))
            self.update(int((z - 1) / length * 100.0))
            yield prop

        logger.debug(notread)



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.
import pytest

# Local modules.
from pyxray.parser.nist import NISTElementAtomicWeightParser

# Globals and constants variables.


def test_coursey2015():
    parser = NISTElementAtomicWeightParser()
    props = list(parser)

    assert len(props) == 84

    assert props[70].element.z == 73
    assert props[70].value == pytest.approx(180.94788, abs=1e-5)



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.

# Local modules.
from pyxray.parser.sargent_welch import (
    SargentWelchElementAtomicWeightParser,
    SargentWelchElementMassDensityParser,
)

# Globals and constants variables.


def test_sargentwelch2010_atomicweight():
    parser = SargentWelchElementAtomicWeightParser()
    assert len(list(parser)) == 106


def test_sargentwelch2010_massdensity():
    parser = SargentWelchElementMassDensityParser()
    assert len(list(parser)) == 94



""""""

__all__ = ["Composition"]

# Standard library modules.
import math
from types import MappingProxyType
import itertools
from fractions import Fraction
import re

# Third party modules.
import pyxray

# Local modules.

# Globals and constants variables.
CHEMICAL_FORMULA_PATTERN = re.compile(r"([A-Z][a-z]?)([0-9\.]*)")


def process_wildcard(fractions):
    """
    Processes element with a wildcard ``?`` weight fraction and returns
    composition balanced to 1.0.
    """
    wildcard_zs = set()
    total_fraction = 0.0
    for z, fraction in fractions.items():
        if fraction == "?":
            wildcard_zs.add(z)
        else:
            total_fraction += fraction

    if not wildcard_zs:
        return fractions

    balance_fraction = (1.0 - total_fraction) / len(wildcard_zs)
    for z in wildcard_zs:
        fractions[z] = balance_fraction

    return fractions


def convert_mass_to_atomic_fractions(mass_fractions):
    """
    Converts a mass fraction :class:`dict` to an atomic fraction :class:`dict`.

    Args:
        mass_fractions (dict): mass fraction :class:`dict`.
            The composition is specified by a dictionary.
            The keys are atomic numbers and the values weight fractions.
            No wildcard are accepted.
    """
    atomic_fractions = {}

    for z, mass_fraction in mass_fractions.items():
        atomic_fractions[z] = mass_fraction / pyxray.element_atomic_weight(z)

    total_fraction = sum(atomic_fractions.values())

    for z, fraction in atomic_fractions.items():
        try:
            atomic_fractions[z] = fraction / total_fraction
        except ZeroDivisionError:
            atomic_fractions[z] = 0.0

    return atomic_fractions


def convert_atomic_to_mass_fractions(atomic_fractions):
    """
    Converts an atomic fraction :class:`dict` to a mass fraction :class:`dict`.

    Args:
        atomic_fractions (dict): atomic fraction :class:`dict`.
            The composition is specified by a dictionary.
            The keys are atomic numbers and the values atomic fractions.
            No wildcard are accepted.
    """
    # Calculate total atomic mass
    atomic_masses = {}
    total_atomic_mass = 0.0
    for z, atomic_fraction in atomic_fractions.items():
        atomic_mass = pyxray.element_atomic_weight(z)
        atomic_masses[z] = atomic_mass
        total_atomic_mass += atomic_fraction * atomic_mass

    # Create mass fraction
    mass_fractions = {}
    for z, atomic_fraction in atomic_fractions.items():
        mass_fractions[z] = atomic_fraction * atomic_masses[z] / total_atomic_mass

    return mass_fractions


def convert_formula_to_atomic_fractions(formula):
    """
    Converts a chemical formula to an atomic fraction :class:`dict`.

    Args:
        formula (str): chemical formula, like Al2O3. No wildcard are accepted.
    """
    mole_fractions = {}
    total_mole_fraction = 0.0

    for match in CHEMICAL_FORMULA_PATTERN.finditer(formula):
        symbol, mole_fraction = match.groups()

        z = pyxray.element_atomic_number(symbol.strip())

        if mole_fraction == "":
            mole_fraction = 1.0
        mole_fraction = float(mole_fraction)

        mole_fraction = float(mole_fraction)
        mole_fractions[z] = mole_fraction
        total_mole_fraction += mole_fraction

    # Calculate atomic fractions
    atomic_fractions = {}
    for z, mole_fraction in mole_fractions.items():
        atomic_fractions[z] = mole_fraction / total_mole_fraction

    return atomic_fractions


def generate_name(atomic_fractions):
    """
    Generates a name from the composition.
    The name is generated on the basis of a classical chemical formula.
    """
    if not atomic_fractions:
        return ""

    if len(atomic_fractions) == 1:
        z = list(atomic_fractions.keys())[0]
        return pyxray.element_symbol(z)

    symbols = []
    fractions = []
    for z in sorted(atomic_fractions.keys(), reverse=True):
        symbols.append(pyxray.element_symbol(z))
        fractions.append(Fraction(atomic_fractions[z]).limit_denominator())

    # Find gcd of the fractions
    gcds = []
    for a, b in itertools.combinations(fractions, 2):
        gcds.append(math.gcd(a.denominator, b.denominator))
    smallest_gcd = min(gcds)

    # Write formula
    name = ""
    for symbol, fraction in zip(symbols, fractions):
        mole_fraction = int(fraction * smallest_gcd)
        if mole_fraction == 0:
            continue
        elif mole_fraction == 1:
            name += "%s" % symbol
        else:
            name += "%s%i" % (symbol, mole_fraction)

    return name


class Composition:
    """
    Defines a composition of a compound.

    To create a composition, use the class methods:

        - :meth:`from_pure`
        - :meth:`from_formula`
        - :meth:`from_mass_fractions`
        - :meth:`from_atomic_fractions`

    Use the following attributes to access the composition values:

        - :attr:`mass_fractions`: :class:`dict` where the keys are atomic numbers and the values weight fractions.
        - :attr:`atomic_fractions`: :class:`dict` where the keys are atomic numbers and the values atomic fractions.
        - :attr:`formula`: chemical formula

    The composition object is immutable, i.e. it cannot be modified once created.
    Equality can be checked.
    It is hashable.
    It can be pickled or copied.
    """

    _key = object()
    PRECISION = 0.000000001  # 1ppb

    def __init__(self, key, mass_fractions, atomic_fractions, formula):
        """
        Private constructor. It should never be used.
        """
        if key != Composition._key:
            raise TypeError("Composition cannot be created using constructor")
        if set(mass_fractions.keys()) != set(atomic_fractions.keys()):
            raise ValueError("Mass and atomic fractions must have the same elements")

        self.mass_fractions = MappingProxyType(mass_fractions)
        self.atomic_fractions = MappingProxyType(atomic_fractions)
        self._formula = formula

    @classmethod
    def from_pure(cls, z):
        """
        Creates a pure composition.

        Args:
            z (int): atomic number
        """
        return cls(cls._key, {z: 1.0}, {z: 1.0}, pyxray.element_symbol(z))

    @classmethod
    def from_formula(cls, formula):
        """
        Creates a composition from a chemical formula.

        Args:
            formula (str): chemical formula
        """
        atomic_fractions = convert_formula_to_atomic_fractions(formula)
        return cls.from_atomic_fractions(atomic_fractions)

    @classmethod
    def from_mass_fractions(cls, mass_fractions, formula=None):
        """
        Creates a composition from a mass fraction :class:`dict`.

        Args:
            mass_fractions (dict): mass fraction :class:`dict`.
                The keys are atomic numbers and the values weight fractions.
                Wildcard are accepted, e.g. ``{5: '?', 25: 0.4}`` where boron
                will get a mass fraction of 0.6.
            formula (str): optional chemical formula for the composition.
                If ``None``, a formula will be generated for the composition.
        """
        mass_fractions = process_wildcard(mass_fractions)
        atomic_fractions = convert_mass_to_atomic_fractions(mass_fractions)
        if not formula:
            formula = generate_name(atomic_fractions)
        return cls(cls._key, mass_fractions, atomic_fractions, formula)

    @classmethod
    def from_atomic_fractions(cls, atomic_fractions, formula=None):
        """
        Creates a composition from an atomic fraction :class:`dict`.

        Args:
            atomic_fractions (dict): atomic fraction :class:`dict`.
                The keys are atomic numbers and the values atomic fractions.
                Wildcard are accepted, e.g. ``{5: '?', 25: 0.4}`` where boron
                will get a atomic fraction of 0.6.
            formula (str): optional chemical formula for the composition.
                If ``None``, a formula will be generated for the composition.
        """
        atomic_fractions = process_wildcard(atomic_fractions)
        mass_fractions = convert_atomic_to_mass_fractions(atomic_fractions)
        if not formula:
            formula = generate_name(atomic_fractions)
        return cls(cls._key, mass_fractions, atomic_fractions, formula)

    def __len__(self):
        return len(self.mass_fractions)

    def __contains__(self, z):
        return z in self.mass_fractions

    def __iter__(self):
        return iter(self.mass_fractions.keys())

    def __repr__(self):
        return "<{}({})>".format(self.__class__.__name__, self.inner_repr())

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        if len(self) != len(other):
            return False

        for z in self.mass_fractions:
            if z not in other.mass_fractions:
                return False

            fraction = self.mass_fractions[z]
            other_fraction = other.mass_fractions[z]

            if not math.isclose(fraction, other_fraction, abs_tol=self.PRECISION):
                return False

        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        out = []
        for z in sorted(self.mass_fractions):
            out.append(z)
            out.append(int(self.mass_fractions[z] / self.PRECISION))

        return hash(tuple(out))

    def __getstate__(self):
        return {
            "mass_fractions": dict(self.mass_fractions),
            "atomic_fractions": dict(self.atomic_fractions),
            "formula": self.formula,
        }

    def __setstate__(self, state):
        self.mass_fractions = MappingProxyType(state.get("mass_fractions", {}))
        self.atomic_fractions = MappingProxyType(state.get("atomic_fractions", {}))
        self._formula = state.get("formula", "")

    def is_normalized(self):
        return math.isclose(
            sum(self.mass_fractions.values()), 1.0, abs_tol=self.PRECISION
        )

    def inner_repr(self):
        return ", ".join(
            "{}: {:.4f}".format(pyxray.element_symbol(z), mass_fraction)
            for z, mass_fraction in self.mass_fractions.items()
        )

    @property
    def formula(self):
        return self._formula



#!/usr/bin/env python
""" """

# Standard library modules.
import os
import sqlite3

# Third party modules.
import pytest

# Local modules.

# Globals and constants variables.


def test_database_build(builder):
    assert os.path.exists(builder.engine.url.database)

    conn = sqlite3.connect(builder.engine.url.database)
    command = "SELECT count(*) FROM sqlite_master WHERE type = 'table'"
    (ntable,) = conn.execute(command).fetchone()
    assert ntable == 21


def test_database_fail(builder):
    builder.badparser = True

    with pytest.raises(Exception):
        builder.build()



"""
Build the SQL database from the registered parsers.
"""

# Standard library modules.
import dataclasses
import logging

# Third party modules.
import tqdm

# Local modules.
from pyxray.parser.base import find_parsers
from pyxray.sql.base import SqlBase

# Globals and constants variables.
logger = logging.getLogger(__name__)


class SqlDatabaseBuilder(SqlBase):
    def _convert_dataclass_to_params(self, dataclass):
        params = {}
        for field in dataclasses.fields(dataclass):
            name = field.name
            value = getattr(dataclass, name)

            if dataclasses.is_dataclass(value):
                row_id = self.insert(value, check_duplicate=True)
                params[name + "_id"] = row_id
            else:
                params[name] = value

        return params

    def insert(self, dataclass, check_duplicate=False):
        """
        Inserts a dataclass instance.

        Args:
            dataclass (dataclasses.dataclass): instance
            check_duplicate (bool): whether to check if the dataclass already
                contains a matching row

        Returns:
            int: row of the dataclass instance in its table
        """
        # Check if already exists
        if check_duplicate:
            row_id = self._get_row(dataclass)
            if row_id is not None:
                return row_id

        # Create insert statement
        table = self.require_table(dataclass)
        params = self._convert_dataclass_to_params(dataclass)
        ins = table.insert().values(**params)

        logger.debug('Insert in "{}": {!r}'.format(table.name, params))

        # Insert
        with self.engine.begin() as conn:
            result = conn.execute(ins)
            return result.inserted_primary_key[0]

    def insert_many(self, list_dataclass):
        clasz = type(list_dataclass[0])

        list_params = []
        for dataclass in list_dataclass:
            if type(dataclass) != clasz:
                raise ValueError("All dataclasses do not have the same type")

            params = self._convert_dataclass_to_params(dataclass)
            list_params.append(params)

        table = self.require_table(clasz)

        # Insert
        with self.engine.begin() as conn:
            conn.execute(table.insert(), list_params)

    def _find_parsers(self):
        return find_parsers()

    def build(self):
        """
        Find all parsers and insert their properties in the database.
        """
        parsers = self._find_parsers()
        logger.info("Found {:d} parsers".format(len(parsers)))

        for name, parser in tqdm.tqdm(parsers, desc="Building database"):
            buffer = {}

            for prop in tqdm.tqdm(parser, desc="Processing {}".format(name)):
                buffer.setdefault(type(prop), []).append(prop)

            for list_dataclass in tqdm.tqdm(
                buffer.values(), desc="Inserting {}".format(name)
            ):
                self.insert_many(list_dataclass)



""""""

# Standard library modules.
from collections.abc import Sequence
import logging

# Third party modules.
import sqlalchemy.sql

# Local modules.
from pyxray.base import _DatabaseMixin, NotFound
from pyxray.sql.base import SqlBase
import pyxray.descriptor as descriptor
import pyxray.property as property

# Globals and constants variables.
logger = logging.getLogger(__name__)


class StatementBuilder:
    def __init__(self):
        self._columns = []
        self._joins = {}
        self._clauses = []

    def add_column(self, column):
        self._columns.append(column)

    def add_join(self, left, right, onclause):
        if left == right:
            return
        self._joins[(left, right)] = onclause

    def add_clause(self, clause):
        self._clauses.append(clause)

    def build(self):
        statement = sqlalchemy.sql.select(self._columns)

        if self._joins:
            # Joins have to be nested to work with sqlalchemy
            # E.g.
            # j = table_a.join(
            #        table_b.join(table_c,
            #            table_b.c.id == table_c.c.b_id),
            #        table_b.c.a_id == table_a.c.id)
            joins = self._joins.copy()

            (left, right), onclause = joins.popitem()
            finaljoin = left.join(right, onclause)

            for (_left, right), onclause in joins.items():
                finaljoin = finaljoin.join(right, onclause)

            statement = statement.select_from(finaljoin)

        return statement.where(sqlalchemy.sql.and_(*self._clauses))


class SqlDatabase(_DatabaseMixin, SqlBase):
    def __init__(self, engine):
        super().__init__(engine)
        self._preferred_references = []

    def _expand_atomic_subshell(self, atomic_subshell):
        if (
            hasattr(atomic_subshell, "principal_quantum_number")
            and hasattr(atomic_subshell, "azimuthal_quantum_number")
            and hasattr(atomic_subshell, "total_angular_momentum_nominator")
        ):
            n = atomic_subshell.atomic_shell.principal_quantum_number
            l = atomic_subshell.azimuthal_quantum_number
            j_n = atomic_subshell.total_angular_momentum_nominator

        elif isinstance(atomic_subshell, Sequence) and len(atomic_subshell) == 3:
            n = atomic_subshell[0]
            l = atomic_subshell[1]
            j_n = atomic_subshell[2]

        else:
            raise NotFound("Cannot parse atomic subshell: {}".format(atomic_subshell))

        return n, l, j_n

    def _expand_xray_transition(self, xray_transition):
        if isinstance(xray_transition, descriptor.XrayTransition):
            src_n = xray_transition.source_principal_quantum_number
            src_l = xray_transition.source_azimuthal_quantum_number
            src_j_n = xray_transition.source_total_angular_momentum_nominator
            dst_n = xray_transition.destination_principal_quantum_number
            dst_l = xray_transition.destination_azimuthal_quantum_number
            dst_j_n = xray_transition.destination_total_angular_momentum_nominator

        elif isinstance(xray_transition, Sequence) and len(xray_transition) >= 2:
            src_n, src_l, src_j_n = self._expand_atomic_subshell(xray_transition[0])
            dst_n, dst_l, dst_j_n = self._expand_atomic_subshell(xray_transition[1])

        else:
            raise NotFound("Cannot parse X-ray transition: {}".format(xray_transition))

        return src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n

    def _update_element(self, builder, table, element, column="element_id"):
        if hasattr(element, "atomic_number"):
            element = element.atomic_number

        if isinstance(element, str):
            table_name = self.require_table(property.ElementName)
            builder.add_join(
                table, table_name, table.c[column] == table_name.c["element_id"]
            )
            clause_name = table_name.c["value"] == element

            table_symbol = self.require_table(property.ElementSymbol)
            builder.add_join(
                table, table_symbol, table.c[column] == table_symbol.c["element_id"]
            )
            clause_symbol = table_symbol.c["value"] == element

            builder.add_clause(sqlalchemy.sql.or_(clause_name, clause_symbol))

        elif isinstance(element, int):
            table_element = self.require_table(descriptor.Element)
            builder.add_join(
                table, table_element, table.c[column] == table_element.c["id"]
            )
            builder.add_clause(table_element.c["atomic_number"] == element)

        else:
            raise NotFound("Cannot parse element: {}".format(element))

    def _update_atomic_shell(
        self, builder, table, atomic_shell, column="atomic_shell_id"
    ):
        if hasattr(atomic_shell, "principal_quantum_number"):
            atomic_shell = atomic_shell.principal_quantum_number

        if isinstance(atomic_shell, str):
            table_notation = self.require_table(property.AtomicShellNotation)
            builder.add_join(
                table,
                table_notation,
                table.c[column] == table_notation.c["atomic_shell_id"],
            )
            builder.add_clause(
                sqlalchemy.sql.or_(
                    table_notation.c["ascii"] == atomic_shell,
                    table_notation.c["utf16"] == atomic_shell,
                )
            )

        elif isinstance(atomic_shell, int):
            table_atomic_shell = self.require_table(descriptor.AtomicShell)
            builder.add_join(
                table, table_atomic_shell, table.c[column] == table_atomic_shell.c["id"]
            )
            builder.add_clause(
                table_atomic_shell.c["principal_quantum_number"] == atomic_shell
            )

        else:
            raise NotFound("Cannot parse atomic shell: {}".format(atomic_shell))

    def _update_atomic_subshell(
        self, builder, table, atomic_subshell, column="atomic_subshell_id"
    ):
        if isinstance(atomic_subshell, str):
            table_notation = self.require_table(property.AtomicSubshellNotation)
            builder.add_join(
                table,
                table_notation,
                table.c[column] == table_notation.c["atomic_subshell_id"],
            )
            builder.add_clause(
                sqlalchemy.sql.or_(
                    table_notation.c["ascii"] == atomic_subshell,
                    table_notation.c["utf16"] == atomic_subshell,
                )
            )

        else:
            n, l, j_n = self._expand_atomic_subshell(atomic_subshell)
            table_atomic_subshell = self.require_table(descriptor.AtomicSubshell)
            builder.add_join(
                table,
                table_atomic_subshell,
                table.c[column] == table_atomic_subshell.c["id"],
            )
            builder.add_clause(table_atomic_subshell.c["principal_quantum_number"] == n)
            builder.add_clause(table_atomic_subshell.c["azimuthal_quantum_number"] == l)
            builder.add_clause(
                table_atomic_subshell.c["total_angular_momentum_nominator"] == j_n
            )

    def _update_xray_transition(
        self, builder, table, xray_transition, column="xray_transition_id", search=False
    ):
        if isinstance(xray_transition, str):
            table_notation = self.require_table(property.XrayTransitionNotation)
            builder.add_join(
                table,
                table_notation,
                table.c[column] == table_notation.c["xray_transition_id"],
            )
            builder.add_clause(
                sqlalchemy.sql.or_(
                    table_notation.c["ascii"] == xray_transition,
                    table_notation.c["utf16"] == xray_transition,
                )
            )

        else:
            src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n = self._expand_xray_transition(
                xray_transition
            )
            table_xray_transition = self.require_table(descriptor.XrayTransition)
            builder.add_join(
                table,
                table_xray_transition,
                table.c[column] == table_xray_transition.c["id"],
            )

            if search:

                def create_clause(column, value):
                    if value is not None:
                        return table_xray_transition.c[column] == value
                    else:
                        return table_xray_transition.c[column] != None

                builder.add_clause(
                    create_clause("source_principal_quantum_number", src_n)
                )
                builder.add_clause(
                    create_clause("source_azimuthal_quantum_number", src_l)
                )
                builder.add_clause(
                    create_clause("source_total_angular_momentum_nominator", src_j_n)
                )
                builder.add_clause(
                    create_clause("destination_principal_quantum_number", dst_n)
                )
                builder.add_clause(
                    create_clause("destination_azimuthal_quantum_number", dst_l)
                )
                builder.add_clause(
                    create_clause(
                        "destination_total_angular_momentum_nominator", dst_j_n
                    )
                )

            else:
                builder.add_clause(
                    table_xray_transition.c["source_principal_quantum_number"] == src_n
                )
                builder.add_clause(
                    table_xray_transition.c["source_azimuthal_quantum_number"] == src_l
                )
                builder.add_clause(
                    table_xray_transition.c["source_total_angular_momentum_nominator"]
                    == src_j_n
                )
                builder.add_clause(
                    table_xray_transition.c["destination_principal_quantum_number"]
                    == dst_n
                )
                builder.add_clause(
                    table_xray_transition.c["destination_azimuthal_quantum_number"]
                    == dst_l
                )
                builder.add_clause(
                    table_xray_transition.c[
                        "destination_total_angular_momentum_nominator"
                    ]
                    == dst_j_n
                )

    def _update_reference(self, builder, table, reference, column="reference_id"):
        if isinstance(reference, descriptor.Reference):
            reference = reference.bibtexkey

        table_reference = self.require_table(descriptor.Reference)
        builder.add_column(table_reference.c["bibtexkey"])
        builder.add_join(
            table, table_reference, table.c[column] == table_reference.c["id"]
        )

        if reference:
            builder.add_clause(table_reference.c["bibtexkey"] == reference)

    def _update_language(self, builder, table, language):
        if isinstance(language, descriptor.Language):
            language = language.key

        table_language = self.require_table(descriptor.Language)
        builder.add_join(
            table, table_language, table.c["language_id"] == table_language.c["id"]
        )
        builder.add_clause(table_language.c["key"] == language)

    def _update_notation(self, builder, table, notation):
        if isinstance(notation, descriptor.Notation):
            notation = notation.key

        table_notation = self.require_table(descriptor.Notation)
        builder.add_join(
            table, table_notation, table.c["notation_id"] == table_notation.c["id"]
        )
        builder.add_clause(table_notation.c["key"] == notation)

    def _execute(self, builder, remove_bibtexkey_column=True):
        statement = builder.build()
        logger.debug(statement.compile())

        # Execute
        with self.engine.connect() as conn:
            rows = conn.execute(statement).fetchall()
            if not rows:
                raise NotFound

            # Only one row, no need to check for the preferred references
            elif len(rows) == 1:
                row = rows[0]

            else:
                dictrows = dict((dict(row).get("bibtexkey", None), row) for row in rows)
                row = rows[0]  # fall back
                for bibtexkey in self._preferred_references:
                    if bibtexkey in dictrows:
                        row = dictrows[bibtexkey]
                        break

            row = dict(row)
            if remove_bibtexkey_column:
                row.pop("bibtexkey", None)

            if len(row) == 1:
                return row.popitem()[1]
            else:
                return row.values()

    def _execute_many(self, builder, remove_bibtexkey_column=True):
        statement = builder.build()
        logger.debug(statement.compile())

        # Execute
        with self.engine.connect() as conn:
            rows = conn.execute(statement).fetchall()
            if not rows:
                raise NotFound

            outrows = []
            for row in rows:
                row = dict(row)
                if remove_bibtexkey_column:
                    row.pop("bibtexkey", None)
                outrows.append(list(row.values()))

            return outrows

    def get_preferred_references(self):
        return tuple(self._preferred_references)

    def add_preferred_reference(self, reference):
        if isinstance(reference, descriptor.Reference):
            reference = reference.bibtexkey
        if reference in self._preferred_references:
            return

        table = self.require_table(descriptor.Reference)

        builder = StatementBuilder()
        self._update_reference(builder, table, reference, "id")

        bibtexkey = self._execute(builder, remove_bibtexkey_column=False)

        self._preferred_references.append(bibtexkey)

    def clear_preferred_references(self):
        self._preferred_references.clear()

    def element(self, element):
        table = self.require_table(descriptor.Element)

        builder = StatementBuilder()
        builder.add_column(table.c["atomic_number"])
        self._update_element(builder, table, element, "id")

        atomic_number = self._execute(builder)
        return descriptor.Element(atomic_number)

    def element_atomic_number(self, element):
        table = self.require_table(descriptor.Element)

        builder = StatementBuilder()
        builder.add_column(table.c["atomic_number"])
        self._update_element(builder, table, element, "id")

        return self._execute(builder)

    def element_symbol(self, element, reference=None):
        table = self.require_table(property.ElementSymbol)

        builder = StatementBuilder()
        builder.add_column(table.c["value"])
        self._update_element(builder, table, element)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def element_name(self, element, language="en", reference=None):
        table = self.require_table(property.ElementName)

        builder = StatementBuilder()
        builder.add_column(table.c["value"])
        self._update_element(builder, table, element)
        self._update_language(builder, table, language)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def element_atomic_weight(self, element, reference=None):
        table = self.require_table(property.ElementAtomicWeight)

        builder = StatementBuilder()
        builder.add_column(table.c["value"])
        self._update_element(builder, table, element)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def element_mass_density_kg_per_m3(self, element, reference=None):
        table = self.require_table(property.ElementMassDensity)

        builder = StatementBuilder()
        builder.add_column(table.c["value_kg_per_m3"])
        self._update_element(builder, table, element)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def element_xray_transitions(self, element, xray_transition=None, reference=None):
        table_xray = self.require_table(descriptor.XrayTransition)
        table_probability = self.require_table(property.XrayTransitionProbability)

        builder = StatementBuilder()
        builder.add_column(table_xray.c["source_principal_quantum_number"])
        builder.add_column(table_xray.c["source_azimuthal_quantum_number"])
        builder.add_column(table_xray.c["source_total_angular_momentum_nominator"])
        builder.add_column(table_xray.c["destination_principal_quantum_number"])
        builder.add_column(table_xray.c["destination_azimuthal_quantum_number"])
        builder.add_column(table_xray.c["destination_total_angular_momentum_nominator"])
        builder.add_join(
            table_probability,
            table_xray,
            table_probability.c["xray_transition_id"] == table_xray.c["id"],
        )
        builder.add_clause(table_probability.c["value"] > 0.0)
        self._update_element(builder, table_probability, element)
        self._update_reference(builder, table_probability, reference)
        if xray_transition is not None:
            self._update_xray_transition(
                builder, table_probability, xray_transition, search=True
            )

        transitions = []
        try:
            for src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n in self._execute_many(
                builder
            ):
                transition = descriptor.XrayTransition(
                    src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n
                )
                transitions.append(transition)
        except NotFound:
            logger.info("No transition found for {}".format(element))

        if len(transitions) == 0:
            table_relative_weight = self.require_table(
                property.XrayTransitionRelativeWeight
            )
            builder = StatementBuilder()
            builder.add_column(table_xray.c["source_principal_quantum_number"])
            builder.add_column(table_xray.c["source_azimuthal_quantum_number"])
            builder.add_column(table_xray.c["source_total_angular_momentum_nominator"])
            builder.add_column(table_xray.c["destination_principal_quantum_number"])
            builder.add_column(table_xray.c["destination_azimuthal_quantum_number"])
            builder.add_column(
                table_xray.c["destination_total_angular_momentum_nominator"]
            )
            builder.add_join(
                table_relative_weight,
                table_xray,
                table_relative_weight.c["xray_transition_id"] == table_xray.c["id"],
            )
            builder.add_clause(table_relative_weight.c["value"] > 0.0)
            self._update_element(builder, table_relative_weight, element)
            self._update_reference(builder, table_relative_weight, reference)
            if xray_transition is not None:
                self._update_xray_transition(
                    builder, table_relative_weight, xray_transition, search=True
                )

            transitions = []
            try:
                for src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n in self._execute_many(
                    builder
                ):
                    transition = descriptor.XrayTransition(
                        src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n
                    )
                    transitions.append(transition)
            except NotFound:
                logger.info("No transition found for {}".format(element))

        if len(transitions) == 0:
            raise NotFound

        return tuple(transitions)

    def element_xray_transition(self, element, xray_transition, reference=None):
        table_xray = self.require_table(descriptor.XrayTransition)
        table_probability = self.require_table(property.XrayTransitionProbability)

        builder = StatementBuilder()
        builder.add_column(table_xray.c["source_principal_quantum_number"])
        builder.add_column(table_xray.c["source_azimuthal_quantum_number"])
        builder.add_column(table_xray.c["source_total_angular_momentum_nominator"])
        builder.add_column(table_xray.c["destination_principal_quantum_number"])
        builder.add_column(table_xray.c["destination_azimuthal_quantum_number"])
        builder.add_column(table_xray.c["destination_total_angular_momentum_nominator"])
        builder.add_join(
            table_probability,
            table_xray,
            table_probability.c["xray_transition_id"] == table_xray.c["id"],
        )
        builder.add_clause(table_probability.c["value"] > 0.0)
        self._update_xray_transition(builder, table_xray, xray_transition, "id")
        self._update_element(builder, table_probability, element)
        self._update_reference(builder, table_probability, reference)

        src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n = self._execute(builder)
        return descriptor.XrayTransition(src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n)

    def atomic_shell(self, atomic_shell):
        table = self.require_table(descriptor.AtomicShell)

        builder = StatementBuilder()
        builder.add_column(table.c["principal_quantum_number"])
        self._update_atomic_shell(builder, table, atomic_shell, "id")

        principal_quantum_number = self._execute(builder)
        return descriptor.AtomicShell(principal_quantum_number)

    def atomic_shell_notation(
        self, atomic_shell, notation, encoding="utf16", reference=None
    ):
        table = self.require_table(property.AtomicShellNotation)

        builder = StatementBuilder()
        builder.add_column(table.c[encoding])
        self._update_atomic_shell(builder, table, atomic_shell, "id")
        self._update_notation(builder, table, notation)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def atomic_subshell(self, atomic_subshell):
        table = self.require_table(descriptor.AtomicSubshell)

        builder = StatementBuilder()
        builder.add_column(table.c["principal_quantum_number"])
        builder.add_column(table.c["azimuthal_quantum_number"])
        builder.add_column(table.c["total_angular_momentum_nominator"])
        self._update_atomic_subshell(builder, table, atomic_subshell, "id")

        n, l, j_n = self._execute(builder)
        return descriptor.AtomicSubshell(n, l, j_n)

    def atomic_subshell_notation(
        self, atomic_subshell, notation, encoding="utf16", reference=None
    ):
        table = self.require_table(property.AtomicSubshellNotation)

        builder = StatementBuilder()
        builder.add_column(table.c[encoding])
        self._update_atomic_subshell(builder, table, atomic_subshell)
        self._update_notation(builder, table, notation)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def atomic_subshell_binding_energy_eV(
        self, element, atomic_subshell, reference=None
    ):
        table = self.require_table(property.AtomicSubshellBindingEnergy)

        builder = StatementBuilder()
        builder.add_column(table.c["value_eV"])
        self._update_element(builder, table, element)
        self._update_atomic_subshell(builder, table, atomic_subshell)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def atomic_subshell_radiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):
        table = self.require_table(property.AtomicSubshellRadiativeWidth)

        builder = StatementBuilder()
        builder.add_column(table.c["value_eV"])
        self._update_element(builder, table, element)
        self._update_atomic_subshell(builder, table, atomic_subshell)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def atomic_subshell_nonradiative_width_eV(
        self, element, atomic_subshell, reference=None
    ):
        table = self.require_table(property.AtomicSubshellNonRadiativeWidth)

        builder = StatementBuilder()
        builder.add_column(table.c["value_eV"])
        self._update_element(builder, table, element)
        self._update_atomic_subshell(builder, table, atomic_subshell)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def atomic_subshell_occupancy(self, element, atomic_subshell, reference=None):
        table = self.require_table(property.AtomicSubshellOccupancy)

        builder = StatementBuilder()
        builder.add_column(table.c["value"])
        self._update_element(builder, table, element)
        self._update_atomic_subshell(builder, table, atomic_subshell)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def xray_transition(self, xray_transition):
        table = self.require_table(descriptor.XrayTransition)

        builder = StatementBuilder()
        builder.add_column(table.c["source_principal_quantum_number"])
        builder.add_column(table.c["source_azimuthal_quantum_number"])
        builder.add_column(table.c["source_total_angular_momentum_nominator"])
        builder.add_column(table.c["destination_principal_quantum_number"])
        builder.add_column(table.c["destination_azimuthal_quantum_number"])
        builder.add_column(table.c["destination_total_angular_momentum_nominator"])
        self._update_xray_transition(builder, table, xray_transition, "id")

        src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n = self._execute(builder)
        return descriptor.XrayTransition(src_n, src_l, src_j_n, dst_n, dst_l, dst_j_n)

    def xray_transition_notation(
        self, xray_transition, notation, encoding="utf16", reference=None
    ):
        table = self.require_table(property.XrayTransitionNotation)

        builder = StatementBuilder()
        builder.add_column(table.c[encoding])
        self._update_xray_transition(builder, table, xray_transition)
        self._update_notation(builder, table, notation)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def xray_transition_energy_eV(self, element, xray_transition, reference=None):
        table = self.require_table(property.XrayTransitionEnergy)

        builder = StatementBuilder()
        builder.add_column(table.c["value_eV"])
        self._update_element(builder, table, element)
        self._update_xray_transition(builder, table, xray_transition)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def xray_transition_probability(self, element, xray_transition, reference=None):
        table = self.require_table(property.XrayTransitionProbability)

        builder = StatementBuilder()
        builder.add_column(table.c["value"])
        self._update_element(builder, table, element)
        self._update_xray_transition(builder, table, xray_transition)
        self._update_reference(builder, table, reference)

        return self._execute(builder)

    def xray_transition_relative_weight(self, element, xray_transition, reference=None):
        table = self.require_table(property.XrayTransitionRelativeWeight)

        builder = StatementBuilder()
        builder.add_column(table.c["value"])
        self._update_element(builder, table, element)
        self._update_xray_transition(builder, table, xray_transition)
        self._update_reference(builder, table, reference)

        return self._execute(builder)



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.
import pytest

# Local modules.
from pyxray.descriptor import Reference, Element
from pyxray.property import ElementSymbol

# Globals and constants variables.


@pytest.fixture
def reference():
    return Reference("test2016")


@pytest.fixture
def element_symbol(reference):
    return ElementSymbol(reference, Element(6), "C")


def test_element_symbol(element_symbol, reference):
    assert element_symbol.reference == reference
    assert element_symbol.element == Element(6)
    assert element_symbol.value == "C"


def test_element_symbol_validate(reference):
    with pytest.raises(ValueError):
        ElementSymbol(reference, Element(6), "")
        ElementSymbol(reference, Element(6), "CCC")
        ElementSymbol(reference, Element(6), "c")



#!/usr/bin/env python

# Standard library modules.
from pathlib import Path
import logging

# Third party modules.
from setuptools import setup, find_packages
import setuptools.command.build_py as _build_py

# Local modules.
import versioneer

# Globals and constants variables.
BASEDIR = Path(__file__).parent.resolve()
logger = logging.getLogger("pyxray")


class build_py(_build_py.build_py):
    def run(self):
        # Build SQL database
        import sqlalchemy
        import pyxray.sql.build

        logging.basicConfig()
        logger.setLevel(logging.INFO)

        filepath = BASEDIR.joinpath("pyxray", "data", "pyxray.db").resolve()
        if filepath.exists():
            filepath.unlink()

        engine = sqlalchemy.create_engine("sqlite:///" + str(filepath))
        builder = pyxray.sql.build.SqlDatabaseBuilder(engine)
        builder.build()

        try:
            del self.data_files  # Force reinitialization of files to copy
        except AttributeError:
            pass

        super().run()


with open(BASEDIR.joinpath("README.rst"), "r") as fp:
    LONG_DESCRIPTION = fp.read()

PACKAGES = find_packages()

PACKAGE_DATA = {"pyxray": ["data/pyxray.db"]}

with open(BASEDIR.joinpath("requirements.txt"), "r") as fp:
    INSTALL_REQUIRES = fp.read().splitlines()

EXTRAS_REQUIRE = {}
with open(BASEDIR.joinpath("requirements-dev.txt"), "r") as fp:
    EXTRAS_REQUIRE["dev"] = fp.read().splitlines()
with open(BASEDIR.joinpath("requirements-test.txt"), "r") as fp:
    EXTRAS_REQUIRE["test"] = fp.read().splitlines()

CMDCLASS = versioneer.get_cmdclass()
CMDCLASS["build_py"] = build_py

ENTRY_POINTS = {
    "pyxray.parser": [
        "element symbol = pyxray.parser.notation:ElementSymbolParser",
        "atomic shell notation = pyxray.parser.notation:AtomicShellNotationParser",
        "atomic subshell notation = pyxray.parser.notation:AtomicSubshellNotationParser",
        "generic x-ray transition notation = pyxray.parser.notation:GenericXrayTransitionNotationParser",
        "known x-ray transition notation = pyxray.parser.notation:KnownXrayTransitionNotationParser",
        "series x-ray transition notation = pyxray.parser.notation:SeriesXrayTransitionNotationParser",
        "family x-ray transition notation = pyxray.parser.notation:FamilyXrayTransitionNotationParser",
        "wikipedia element name = pyxray.parser.wikipedia:WikipediaElementNameParser",
        "sargent-welch element atomic weight = pyxray.parser.sargent_welch:SargentWelchElementAtomicWeightParser",
        "sargent-welch element mass density = pyxray.parser.sargent_welch:SargentWelchElementMassDensityParser",
        "perkins1991 = pyxray.parser.perkins1991:Perkins1991Parser",
        "nist atomic weight = pyxray.parser.nist:NISTElementAtomicWeightParser",
        "jeol transition = pyxray.parser.jeol:JEOLTransitionParser",
        "campbell2001 = pyxray.parser.campbell2001:CampbellAtomicSubshellRadiativeWidthParser",
        "dtsa1992 subshell = pyxray.parser.dtsa:DtsaSubshellParser",
        "dtsa1992 transition = pyxray.parser.dtsa:DtsaLineParser",
    ],
}

setup(
    name="pyxray",
    version=versioneer.get_version(),
    url="https://github.com/openmicroanalysis/pyxray",
    author="Philippe T. Pinard",
    author_email="philippe.pinard@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    description="Definitions and properties of X-ray transitions",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="MIT license",
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    cmdclass=CMDCLASS,
    entry_points=ENTRY_POINTS,
)



""""""

# Standard library modules.

# Third party modules.
import pytest
import sqlalchemy

# Local modules.
from pyxray.sql.build import SqlDatabaseBuilder
from pyxray.parser.base import _Parser
import pyxray.descriptor as descriptor
import pyxray.property as property

# Globals and constants variables.
K = descriptor.AtomicSubshell(1, 0, 1)
L3 = descriptor.AtomicSubshell(2, 1, 3)
L2 = descriptor.AtomicSubshell(2, 1, 1)


class MockParser(_Parser):
    def __iter__(self):
        reference = descriptor.Reference("lee1966")
        reference2 = descriptor.Reference("doe2016")
        element = descriptor.Element(118)
        atomic_shell = descriptor.AtomicShell(1)
        transition = descriptor.XrayTransition(L3, K)
        transition2 = descriptor.XrayTransition(L2, K)
        transitionset = descriptor.XrayTransition(2, 1, None, K)
        notation = descriptor.Notation("mock")
        notation_iupac = descriptor.Notation("iupac")
        language = descriptor.Language("en")

        yield property.ElementName(reference, element, language, "Vibranium")
        yield property.ElementSymbol(reference, element, "Vi")
        yield property.ElementAtomicWeight(reference, element, 999.1)
        yield property.ElementAtomicWeight(reference2, element, 111.1)
        yield property.ElementMassDensity(reference, element, 999.2)

        yield property.AtomicShellNotation(
            reference, atomic_shell, notation, "a", "b", "c", "d"
        )

        yield property.AtomicSubshellNotation(
            reference, K, notation, "a", "b", "c", "d"
        )
        yield property.AtomicSubshellBindingEnergy(reference, element, K, 0.1)
        yield property.AtomicSubshellRadiativeWidth(reference, element, K, 0.01)
        yield property.AtomicSubshellNonRadiativeWidth(reference, element, K, 0.001)
        yield property.AtomicSubshellOccupancy(reference, element, K, 1)

        yield property.XrayTransitionNotation(
            reference, transition, notation, "a", "b", "c", "d"
        )
        yield property.XrayTransitionNotation(
            reference, transition, notation_iupac, "aa", "bb", "cc", "dd"
        )
        yield property.XrayTransitionEnergy(reference, element, transition, 0.2)
        yield property.XrayTransitionProbability(reference, element, transition, 0.02)
        yield property.XrayTransitionRelativeWeight(
            reference, element, transition, 0.002
        )

        yield property.XrayTransitionNotation(
            reference, transition2, notation, "e", "f", "g", "h"
        )
        yield property.XrayTransitionEnergy(reference, element, transition2, 0.4)
        yield property.XrayTransitionProbability(reference, element, transition2, 0.04)
        yield property.XrayTransitionRelativeWeight(
            reference, element, transition2, 0.004
        )

        yield property.XrayTransitionNotation(
            reference, transitionset, notation, "i", "j", "k", "l"
        )
        yield property.XrayTransitionEnergy(reference, element, transitionset, 0.6)
        yield property.XrayTransitionProbability(
            reference, element, transitionset, 0.06
        )
        yield property.XrayTransitionRelativeWeight(
            reference, element, transitionset, 0.006
        )


class MockBadParser(_Parser):
    def __iter__(self):
        raise Exception


class MockSqliteDatabaseBuilder(SqlDatabaseBuilder):
    def __init__(self, filepath=None, badparser=False):
        super().__init__(filepath)
        self.badparser = badparser

    def _find_parsers(self):
        super()._find_parsers()  # Ignore output, only for coverage

        if self.badparser:
            return [("bad", MockBadParser())]
        else:
            return [("mock", MockParser())]


@pytest.fixture(scope="session")
def builder(tmp_path_factory):
    engine = sqlalchemy.create_engine(
        "sqlite:///" + str(tmp_path_factory.mktemp("test").joinpath("pyxray.sql"))
    )

    builder = MockSqliteDatabaseBuilder(engine)
    builder.build()

    return builder



"""
Cookbook examples and patterns.
"""

# Standard library modules.

# Third party modules.

# Local modules.

# Globals and constants variables.


class ProgressMixin:
    def update(self, progress):
        """
        Update the progress, a value between 0 and 100.
        """
        assert 0 <= progress <= 100
        self._progress = int(progress)

    @property
    def progress(self):
        """
        Current progress, a value between 0 and 100.
        """
        return getattr(self, "_progress", 0)


class ProgressReportMixin(ProgressMixin):
    def add_reporthook(self, hook):
        assert callable(hook)
        if not hasattr(self, "_reporthooks"):
            self._reporthooks = set()
        self._reporthooks.add(hook)

    def clear_reporthooks(self):
        self._reporthooks = set()

    def update(self, progress):
        super().update(progress)
        for hook in getattr(self, "_reporthooks", []):
            hook(progress)


def formatdoc(*formatargs, **formatkwargs):
    def decorate(func):
        func.__doc__ = func.__doc__.format(*formatargs, **formatkwargs)
        return func

    return decorate



#!/usr/bin/env python
""" """

# Standard library modules.

# Third party modules.
import pytest

# Local modules.
from pyxray.parser.dtsa import DtsaSubshellParser, DtsaLineParser

# Globals and constants variables.


def test_dtsa1992_subshell():
    parser = DtsaSubshellParser()
    props = list(parser)

    assert len(props) == 1045

    assert props[0].element.z == 3
    assert props[0].value_eV == pytest.approx(54.75, abs=1e-2)

    assert props[-1].element.z == 95
    assert props[-1].value_eV == pytest.approx(366.49, abs=1e-2)


def test_dtsa1992_line():
    parser = DtsaLineParser()
    props = list(parser)

    assert len(props) == 6434

    assert props[0].element.z == 3
    assert props[0].value_eV == pytest.approx(54.3000, abs=1e-4)

    assert props[1].element.z == 3
    assert props[1].value == pytest.approx(1.00000, abs=1e-5)

    assert props[-2].element.z == 95
    assert props[-2].value_eV == pytest.approx(378.5100, abs=1e-4)

    assert props[-1].element.z == 95
    assert props[-1].value == pytest.approx(0.01000, abs=1e-2)



""""""

# Standard library modules.
import pickle
import copy

# Third party modules.
import pytest

# Local modules.
import pyxray
from pyxray.composition import (
    Composition,
    convert_atomic_to_mass_fractions,
    convert_mass_to_atomic_fractions,
)

# Globals and constants variables.


@pytest.fixture
def al2o3():
    return Composition.from_formula("Al2O3")


def test_convert_al2o3(al2o3):
    atomic_fractions = al2o3.atomic_fractions

    mass_fractions = convert_atomic_to_mass_fractions(atomic_fractions)
    assert mass_fractions[13] == pytest.approx(0.52925, 1e-4)
    assert mass_fractions[8] == pytest.approx(0.47075, 1e-4)

    atomic_fractions2 = convert_mass_to_atomic_fractions(mass_fractions)
    assert atomic_fractions2[13] == pytest.approx(atomic_fractions2[13], 1e-4)
    assert atomic_fractions2[8] == pytest.approx(atomic_fractions2[8], 1e-4)


@pytest.mark.parametrize("z", [5, 26, 92])
def test_from_pure(z):
    comp = Composition.from_pure(z)

    assert comp.mass_fractions[z] == pytest.approx(1.0, 1e-4)
    assert comp.atomic_fractions[z] == pytest.approx(1.0, 1e-4)
    assert comp.formula == pyxray.element_symbol(z)
    assert comp.is_normalized()


@pytest.mark.parametrize("formula", ["Al2Na3B12", "Al2 Na3 B12", "Al2.0 Na3.0 B12.0"])
def test_from_formula(formula):
    comp = Composition.from_formula(formula)

    assert comp.mass_fractions[13] == pytest.approx(0.21358626371988801, 1e-4)
    assert comp.mass_fractions[11] == pytest.approx(0.27298103136883051, 1e-4)
    assert comp.mass_fractions[5] == pytest.approx(0.51343270491128157, 1e-4)

    assert comp.atomic_fractions[13] == pytest.approx(2 / 17, 1e-4)
    assert comp.atomic_fractions[11] == pytest.approx(3 / 17, 1e-4)
    assert comp.atomic_fractions[5] == pytest.approx(12 / 17, 1e-4)

    assert comp.formula == "Al2Na3B12"

    assert comp.is_normalized()


@pytest.mark.parametrize("formula", ["Al", "Al2", "Al3.0"])
def test_from_formula_pure(formula):
    comp = Composition.from_formula(formula)

    assert comp.mass_fractions[13] == pytest.approx(1.0, 1e-4)
    assert comp.atomic_fractions[13] == pytest.approx(1.0, 1e-4)
    assert comp.formula == "Al"
    assert comp.is_normalized()


@pytest.mark.parametrize("z", range(1, 100))
def test_from_formula_symbol(z):
    formula = pyxray.element_symbol(z)
    comp = Composition.from_formula(formula)
    assert comp.formula == formula


def test_from_formula_exception():
    with pytest.raises(Exception):
        Composition.from_formula("Aq2 Na3 B12")


def test_from_massfractions():
    comp = Composition.from_mass_fractions({13: 0.52925, 8: 0.47075})

    assert comp.mass_fractions[13] == pytest.approx(0.52925, 1e-4)
    assert comp.mass_fractions[8] == pytest.approx(0.47075, 1e-4)

    assert comp.atomic_fractions[13] == pytest.approx(0.4, 1e-4)
    assert comp.atomic_fractions[8] == pytest.approx(0.6, 1e-4)

    assert comp.is_normalized()


def test_from_massfractions_unnormalized():
    comp = Composition.from_mass_fractions({13: 0.52925, 8: 0.3})

    assert comp.mass_fractions[13] == pytest.approx(0.52925, 1e-4)
    assert comp.mass_fractions[8] == pytest.approx(0.3, 1e-4)

    assert comp.atomic_fractions[13] == pytest.approx(0.51127, 1e-4)
    assert comp.atomic_fractions[8] == pytest.approx(0.48873, 1e-4)

    assert not comp.is_normalized()


def test_from_massfractions_wildcard():
    comp = Composition.from_mass_fractions({13: 0.52925, 8: "?"})

    assert comp.mass_fractions[13] == pytest.approx(0.52925, 1e-4)
    assert comp.mass_fractions[8] == pytest.approx(0.47075, 1e-4)

    assert comp.atomic_fractions[13] == pytest.approx(0.4, 1e-4)
    assert comp.atomic_fractions[8] == pytest.approx(0.6, 1e-4)

    assert comp.is_normalized()


def test_eq(al2o3):
    assert al2o3 == Composition.from_atomic_fractions({13: 0.4, 8: 0.6})


def test_ne(al2o3):
    assert al2o3 != Composition.from_mass_fractions({13: 0.52925, 8: 0.47075})


def test_hash1(al2o3):
    assert hash(al2o3) == hash(Composition.from_atomic_fractions({13: 0.4, 8: 0.6}))


def test_hash2(al2o3):
    assert hash(al2o3) != hash(
        Composition.from_mass_fractions({13: 0.52925, 8: 0.47075})
    )


def test_mass_fractions(al2o3):
    with pytest.raises(TypeError):
        al2o3.mass_fractions[13] = 0.5


def test_atomic_fractions(al2o3):
    with pytest.raises(TypeError):
        al2o3.atomic_fractions[13] = 0.5


def test_formula(al2o3):
    assert al2o3.formula == "Al2O3"

    comp = Composition.from_atomic_fractions(al2o3.atomic_fractions)
    assert comp.formula == "Al2O3"

    comp = Composition.from_mass_fractions(al2o3.mass_fractions)
    assert comp.formula == "Al2O3"


def test__contains__(al2o3):
    assert 13 in al2o3
    assert 8 in al2o3


def test_inner_repr(al2o3):
    assert "Al: 0.5293, O: 0.4707" == al2o3.inner_repr()


def test__init__():
    with pytest.raises(TypeError):
        Composition(None, {}, {}, "foo")


def test_pickle(al2o3):
    s = pickle.dumps(al2o3)
    assert al2o3 == pickle.loads(s)


def test_copy(al2o3):
    assert al2o3 == copy.copy(al2o3)


def test_deepcopy(al2o3):
    assert al2o3 == copy.deepcopy(al2o3)



"""
Base SQL.
"""

# Standard library modules.
import re
import dataclasses
import inspect
import logging

# Third party modules.
import sqlalchemy.sql

# Local modules.

# Globals and constants variables.
logger = logging.getLogger(__name__)


def camelcase_to_words(text):
    return re.sub("([a-z0-9])([A-Z])", r"\1 \2", text)


class SqlBase:

    FIELDS_TO_SQLTYPE = {
        int: sqlalchemy.Integer,
        float: sqlalchemy.Float,
        str: sqlalchemy.String,
    }

    def __init__(self, engine):
        self.engine = engine
        self.metadata = sqlalchemy.MetaData()

    def _get_table_name(self, dataclass):
        """
        Creates a table name from a dataclass class or instance.

        Args:
            dataclass (dataclasses.dataclass): class or instance

        Returns:
            str: name of table
        """
        if not inspect.isclass(dataclass):
            dataclass = type(dataclass)
        return "_".join(camelcase_to_words(dataclass.__name__).split()).lower()

    def _create_table(self, table_name, dataclass):
        """
        Creates a table in the database.

        Args:
            table_name (str): name of table
            dataclass (dataclasses.dataclass): class or instance

        Returns:
            :class:`sqlalchemy.Table`: table instance
        """
        columns = [sqlalchemy.Column("id", sqlalchemy.Integer, primary_key=True)]

        for field in dataclasses.fields(dataclass):
            if dataclasses.is_dataclass(field.type):
                subtable = self.require_table(field.type)
                column = sqlalchemy.Column(
                    field.name + "_id",
                    None,
                    sqlalchemy.ForeignKey(subtable.name + ".id"),
                )

            elif field.type in self.FIELDS_TO_SQLTYPE:
                nullable = field.default is None

                if field.type == str and (
                    field.name.startswith("key") or field.name.endswith("key")
                ):
                    columntype = sqlalchemy.String(collation="NOCASE")
                else:
                    columntype = self.FIELDS_TO_SQLTYPE[field.type]

                column = sqlalchemy.Column(field.name, columntype, nullable=nullable)

            else:
                raise ValueError("Unknown field: {}".format(field.type))

            columns.append(column)

        table = sqlalchemy.Table(table_name, self.metadata, *columns)

        self.metadata.create_all(self.engine, tables=[table])
        logger.debug('Create table "{}"'.format(table_name))

        return table

    def require_table(self, dataclass):
        """
        Returns the table for the specified dataclass.
        If no table exists, it is created first.

        Args:
            dataclass (dataclasses.dataclass): class or instance

        Returns:
            :class:`sqlalchemy.Table`: table instance
        """
        table_name = self._get_table_name(dataclass)
        table = self.metadata.tables.get(table_name)

        if table is None:
            table = self._create_table(table_name, dataclass)

        return table

    def _get_row(self, dataclass):
        """
        Returns the row of the dataclass if it exists.
        If not, ``None`` is returned

        Args:
            dataclass (dataclasses.dataclass): instance

        Returns:
            int: row of the dataclass instance in its table, ``None`` if not found
        """
        table = self.require_table(dataclass)

        clauses = []
        for field in dataclasses.fields(dataclass):
            name = field.name
            value = getattr(dataclass, name)

            if dataclasses.is_dataclass(field.type):
                row_id = self._get_row(value)
                clause = table.c[name + "_id"] == row_id

            else:
                clause = table.c[name] == value

            clauses.append(clause)

        statement = sqlalchemy.sql.select([table.c.id]).where(
            sqlalchemy.sql.and_(*clauses)
        )

        with self.engine.begin() as conn:
            return conn.execute(statement).scalar()
